==> nono-0.7.0.tar.gz: nono-0.7.0/vm/virtio_base.h <==
//
// nono
// Copyright (C) 2024 nono project
// Licensed under nono-license.txt
//

//
// VirtIO
//

#pragma once

#include "device.h"
#include "message.h"
#include "thread.h"
#include <array>
#include <condition_variable>

class InterruptDevice;
class MainbusDevice;
class TextScreen;
class VirtIOThread;
class VirtQDesc;

class VirtQueue
{
 public:
	VirtQueue(int idx_, const std::string& name_, int num_max_);

	uint32 last_avail_idx {};

	int idx {};				// 自身が何番目か
	std::string name {};	// キュー名(仕様から決まる)

	uint32 desc {};			// QUEUE_DESC (下位32ビットのみ)
	uint32 driver {};		// QUEUE_DRIVER (下位32ビットのみ)
	uint32 device {};		// QUEUE_DEVICE (下位32ビットのみ)

	int num {};				// QUEUE_NUM (合意できたディスクリプタ数)
	int num_max {};			// QUEUE_NUM の上限
	uint32 ready {};		// QUEUE_READY
};

class VirtIOSeg
{
 public:
	uint32 addr {};			// バッファアドレス (ゲストアドレス)
	uint32 len {};			// バッファ長

	VirtIOSeg(uint32 addr_, uint32 len_)
		: addr(addr_), len(len_)
	{
	}
};

// ディスクリプタチェインの基本クラス
class VirtIOReq
{
 public:
	VirtQueue *q {};
	uint32 idx {};			// このディスクリプタチェインの先頭インデックス
	uint32 len {};			// このディスクリプタチェインの総バイト数
	std::vector<VirtIOSeg> buf {};	// バッファセグメント

	int curseg {};			// 書き込み位置、現在のセグメント
	int offset {};			// 書き込み位置、セグメント内のオフセット
};

// VirtIO 基本クラス
class VirtIODevice : public IODevice
{
	using inherited = IODevice;

 protected:
	// 仕様書には名前が出てこないけど上限
	static const uint MAX_FEATURES_SEL = 2;

	// モニタ幅
	static const int MONITOR_WIDTH = 75;

	// 動作には不要だがモニタに表示したいので…。
	static const uint32 baseaddr = 0xff070000;

 public:
	VirtIODevice(int objid_, int slot_);
	~VirtIODevice() override;

	bool Create() override;
	void SetLogLevel(int loglevel_) override;
	bool Init() override;
	void ResetHard(bool poweron) override;

	busdata Read8(uint32 addr) override;
	busdata Read16(uint32 addr) override;
	busdata Read32(uint32 addr) override;
	busdata Write8(uint32 addr, uint32 data) override;
	busdata Write16(uint32 addr, uint32 data) override;
	busdata Write32(uint32 addr, uint32 data) override;
	busdata Peek8(uint32 addr) override;

	// QUEUE_NOTIFY の処理 (裏スレッドから呼ばれる)
	void QueueNotify(int idx);

 protected:
	int MonitorUpdateDev(TextScreen&, int y) const;
	int MonitorUpdateVirtQueue(TextScreen&, int y, const VirtQueue&) const;
	int MonitorUpdateVirtQDesc(TextScreen&, int y, const VirtQueue&) const;
	virtual uint32 GetReg(uint32 addr) const;
	virtual void QueueReady();

	// ディスクリプタを一つ処理する。
	virtual void ProcessDesc(VirtQueue *) = 0;

	// ディスクリプタを更新する。
	void CommitDesc(VirtQueue *, uint32 idx, uint32 len);

	// 裏スレッドからの処理完了メッセージ
	void DoneMessage(MessageID, uint32);
	// 処理完了
	void Done(VirtQueue *);

	void ChangeInterrupt();

	// メインバスアクセス
	uint32 ReadU8(uint32 addr);
	uint32 ReadLE16(uint32 addr);
	uint32 ReadLE32(uint32 addr);
	uint64 ReadLE64(uint32 addr);
	void WriteU8(uint32 addr, uint32 data);
	void WriteLE16(uint32 addr, uint32 data);
	void WriteLE32(uint32 addr, uint32 data);
	void WriteLE64(uint32 addr, uint64 data);
	uint32 PeekLE16(uint32 addr) const;
	uint32 PeekLE32(uint32 addr) const;

	uint32 ReadDesc(const VirtQueue *, VirtQDesc *desc, uint32 idx);
	void AddDataSegment(VirtIOReq *, uint32 addr, uint32 len, const char *);
	bool ReqAppendByte(VirtIOReq *, uint32 data);

	// DEVICE_FEATURES の feature 番目のビットを立てる。初期化用。
	void SetDeviceFeatures(int feature);
	// DEVICE_FEATURES の feature 番目の名前を返す。
	virtual const char *GetFeatureName(int feature) const;

	// DRIVER_FEATURES の feature 番目のビットが立っていれば true。
	bool GetDriverFeatures(int feature) const;

	int slot {};				// VirtIO のスロット番号 (0-127)

	uint32 device_id {};		// DEVICE_ID
	uint32 vendor_id {};		// VENDOR_ID
	uint32 device_status {};	// STATUS

	uint32 device_sel {};		// DEVICE_FEATURES_SEL
	uint32 driver_sel {};		// DRIVER_FEATURES_SEL
	std::array<uint32, 2> device_features {};	// DEVICE_FEATURES
	std::array<uint32, 2> driver_features {};	// DRIVER_FEATURES

	uint32 queue_num {};		// QUEUE_NUM

	// VirtQueue。継承先で個数が決まっている。
	std::vector<VirtQueue> vqueues {};
	// QUEUE_SEL の値。
	uint32 queue_sel {};
	// QUEUE_SEL で選択されている現在の VirtQueue。選択されてなければ NULL。
	VirtQueue *vq {};

	uint32 intr_status {};			// INTERRUPT_STATUS

	// 0x100 以降の領域
	std::array<uint8, 0x100> device_config {};

	char intrname[10] {};			// 割り込み名

	MessageID msgid {};				// 完了通知のメッセージ ID

	std::unique_ptr<VirtIOThread> backend /*{}*/;

	InterruptDevice *interrupt {};
	MainbusDevice *mainbus {};
};

// VirtIO 空きスロット用
class VirtIONoneDevice : public IODevice
{
	using inherited = IODevice;
 public:
	VirtIONoneDevice();
	~VirtIONoneDevice() override;

	busdata Read32(uint32 addr) override;
	busdata Write32(uint32 addr, uint32 data) override;
	busdata Peek8(uint32 addr) override;
};

// 裏スレッド
class VirtIOThread : public ThreadDevice
{
	using inherited = ThreadDevice;

 protected:
	// REQ_QUEUE はキュー1本に1ビットずつ割り当てているので現状最大31まで。
	static const uint32 REQ_TERMINATE	= 0x80000000;	// スレッド終了指示
	static constexpr uint32 REQ_QUEUE(int idx) {		// QUEUE_NOTIFY 指示
		return (1U << idx);
	}

 public:
	VirtIOThread(VirtIODevice *parent_);
	~VirtIOThread() override;

	void Terminate() override;

	// ゲストからの QUEUE_NOTIFY 要求
	void RequestQueueNotify(int);

 private:
	void ThreadRun() override;

 protected:
	// リクエストフラグ
	uint32 request {};
	std::mutex mtx {};
	std::condition_variable cv {};

	VirtIODevice *parent {};
};

==> nono-0.7.2.tar.gz: nono-0.7.2/vm/virtio_base.h <==
//
// nono
// Copyright (C) 2024 nono project
// Licensed under nono-license.txt
//

//
// VirtIO
//

#pragma once

#include "device.h"
#include "message.h"
#include "thread.h"
#include <array>
#include <condition_variable>

class InterruptDevice;
class MainbusDevice;
class MainRAMDevice;
class TextScreen;
class VirtIOThread;
class VirtQDesc;

class VirtQueue
{
 public:
	VirtQueue(int idx_, const std::string& name_, int num_max_);

	uint32 last_avail_idx {};

	int idx {};				// 自身が何番目か
	std::string name {};	// キュー名(仕様から決まる)

	uint32 desc {};			// QUEUE_DESC (下位32ビットのみ)
	uint32 driver {};		// QUEUE_DRIVER (下位32ビットのみ)
	uint32 device {};		// QUEUE_DEVICE (下位32ビットのみ)

	int num {};				// QUEUE_NUM (合意できたディスクリプタ数)
	int num_max {};			// QUEUE_NUM の上限
	uint32 ready {};		// QUEUE_READY
};

class VirtIOSeg
{
 public:
	uint32 addr {};			// セグメントアドレス (ゲストアドレス)
	uint32 pos {};			// このセグメントのチェイン先頭からの位置
	uint32 len {};			// このセグメント長

	VirtIOSeg(uint32 addr_, uint32 pos_, uint32 len_)
		: addr(addr_), pos(pos_), len(len_)
	{
	}
};

// ディスクリプタチェイン。
// ファイルハンドルのようなものも兼ねている。
// 元の仕様からシーケンシャルアクセスを想定しているようなので、ここでも
// シーケンシャルアクセスを前提とする。そのため pos が UsedRing に書き込む
// バイト数になる。
class VirtIOReq
{
 public:
	VirtQueue *q {};
	uint32 idx {};			// このディスクリプタチェインの先頭インデックス
	uint32 len {};			// このディスクリプタチェインの総バイト数
	uint32 pos {};			// 現在位置 (Used としても使用する)
	std::vector<VirtIOSeg> buf {};	// バッファセグメント

	int curseg {};			// 現在のセグメント (buf[] のインデックス)
};

// VirtIO 基本クラス
class VirtIODevice : public IODevice
{
	using inherited = IODevice;

 protected:
	// 仕様書には名前が出てこないけど上限
	static const uint MAX_FEATURES_SEL = 2;

	// モニタ幅
	static const int MONITOR_WIDTH = 75;

	// 動作には不要だがモニタに表示したいので…。
	static const uint32 baseaddr = 0xff070000;

 public:
	VirtIODevice(int objid_, int slot_);
	~VirtIODevice() override;

	bool Create() override;
	void SetLogLevel(int loglevel_) override;
	bool Init() override;
	void ResetHard(bool poweron) override;

	busdata Read8(uint32 addr) override;
	busdata Read16(uint32 addr) override;
	busdata Read32(uint32 addr) override;
	busdata Write8(uint32 addr, uint32 data) override;
	busdata Write16(uint32 addr, uint32 data) override;
	busdata Write32(uint32 addr, uint32 data) override;
	busdata Peek8(uint32 addr) override;

	// QUEUE_NOTIFY の処理 (裏スレッドから呼ばれる)
	void QueueNotify(int idx);

 protected:
	int MonitorUpdateDev(TextScreen&, int y) const;
	int MonitorUpdateVirtQueue(TextScreen&, int y, const VirtQueue&) const;
	int MonitorUpdateVirtQDesc(TextScreen&, int y, const VirtQueue&) const;
	virtual uint32 GetReg(uint32 addr) const;
	virtual void QueueReady();

	// ディスクリプタを一つ処理する。
	virtual void ProcessDesc(VirtIOReq&) = 0;

	// ディスクリプタを読んでチェインを用意する。
	void StartDesc(VirtIOReq&);
	// ディスクリプタを更新する。
	void CommitDesc(VirtIOReq&);

	// ディスクリプタアクセス
	uint32 ReqReadU8(VirtIOReq&);
	uint32 ReqReadLE16(VirtIOReq&);
	uint64 ReqReadLE32(VirtIOReq&);
	bool ReqReadLE32(VirtIOReq&, uint32 *);
	bool ReqReadLE64(VirtIOReq&, uint64 *);
	bool ReqWriteU8(VirtIOReq&, uint32 data);

	// 裏スレッドからの処理完了メッセージ
	void DoneMessage(MessageID, uint32);
	// 処理完了
	void Done(VirtQueue *);

	void ChangeInterrupt();

	// メインバスアクセス
	uint32 ReadU8(uint32 addr);
	uint32 ReadLE16(uint32 addr);
	uint32 ReadLE32(uint32 addr);
	uint64 ReadLE64(uint32 addr);
	void WriteU8(uint32 addr, uint32 data);
	void WriteLE16(uint32 addr, uint32 data);
	void WriteLE32(uint32 addr, uint32 data);
	void WriteLE64(uint32 addr, uint64 data);
	uint32 PeekLE16(uint32 addr) const;
	uint32 PeekLE32(uint32 addr) const;

	// DEVICE_FEATURES の feature 番目のビットを立てる。初期化用。
	void SetDeviceFeatures(int feature);
	// DEVICE_FEATURES の feature 番目の名前を返す。
	virtual const char *GetFeatureName(int feature) const;

	// DRIVER_FEATURES の feature 番目のビットが立っていれば true。
	bool GetDriverFeatures(int feature) const;

 private:
	// 間接ディスクリプタを追加する。
	void AddIndirectDesc(VirtIOReq&, uint32 addr, uint32 len);

 protected:
	int slot {};				// VirtIO のスロット番号 (0-127)

	uint32 device_id {};		// DEVICE_ID
	uint32 vendor_id {};		// VENDOR_ID
	uint32 device_status {};	// STATUS

	uint32 device_sel {};		// DEVICE_FEATURES_SEL
	uint32 driver_sel {};		// DRIVER_FEATURES_SEL
	std::array<uint32, 2> device_features {};	// DEVICE_FEATURES
	std::array<uint32, 2> driver_features {};	// DRIVER_FEATURES

	uint32 queue_num {};		// QUEUE_NUM

	// VirtQueue。継承先で個数が決まっている。
	std::vector<VirtQueue> vqueues {};
	// QUEUE_SEL の値。
	uint32 queue_sel {};
	// QUEUE_SEL で選択されている現在の VirtQueue。選択されてなければ NULL。
	VirtQueue *vq {};

	uint32 intr_status {};			// INTERRUPT_STATUS

	// 0x100 以降の領域
	std::array<uint8, 0x100> device_config {};

	char intrname[10] {};			// 割り込み名

	MessageID msgid {};				// 完了通知のメッセージ ID

	std::unique_ptr<VirtIOThread> backend /*{}*/;

	InterruptDevice *interrupt {};
	MainbusDevice *mainbus {};
	MainRAMDevice *mainram {};
};

// VirtIO 空きスロット用
class VirtIONoneDevice : public IODevice
{
	using inherited = IODevice;
 public:
	VirtIONoneDevice();
	~VirtIONoneDevice() override;

	busdata Read32(uint32 addr) override;
	busdata Write32(uint32 addr, uint32 data) override;
	busdata Peek8(uint32 addr) override;
};

// 裏スレッド
class VirtIOThread : public ThreadDevice
{
	using inherited = ThreadDevice;

 protected:
	// REQ_QUEUE はキュー1本に1ビットずつ割り当てているので現状最大31まで。
	static const uint32 REQ_TERMINATE	= 0x80000000;	// スレッド終了指示
	static constexpr uint32 REQ_QUEUE(int idx) {		// QUEUE_NOTIFY 指示
		return (1U << idx);
	}

 public:
	VirtIOThread(VirtIODevice *parent_);
	~VirtIOThread() override;

	void Terminate() override;

	// ゲストからの QUEUE_NOTIFY 要求
	void RequestQueueNotify(int);

 private:
	void ThreadRun() override;

 protected:
	// リクエストフラグ
	uint32 request {};
	std::mutex mtx {};
	std::condition_variable cv {};

	VirtIODevice *parent {};
};
